/*
    PPManagedColor.m

    Copyright 2013-2016 Josh Freeman
    http://www.twilightedge.com

    This file is part of PikoPixel for Mac OS X and GNUstep.
    PikoPixel is a graphical application for drawing & editing pixel-art images.

    PikoPixel is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3,
    as published by the Free Software Foundation.

    PikoPixel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#import "PPManagedColor.h"

#import "NSBitmapImageRep_PPUtilities.h"


#if PP_DEPLOYMENT_TARGET_SUPPORTS_COLOR_MANAGEMENT

#pragma mark Managed Color Supported

@interface NSColorSpace (PPManagedColorPrivateMethods)

+ (NSColorSpace *) ppColorSpace_sRGB;

@end

@interface NSData (PPManagedColorPrivateMethods)

+ (NSData *) ppICCProfileFromSystemLibraryFolder_sRGB;

@end


@implementation NSColor (PPManagedColor)

+ (NSColor *) ppManagedColor_sRGBColorWithRed: (CGFloat) red
                                        green: (CGFloat) green
                                        blue: (CGFloat) blue
                                        alpha: (CGFloat) alpha
{
    CGFloat components[4] = {red, green, blue, alpha};

    return [self colorWithColorSpace: [NSColorSpace ppColorSpace_sRGB]
                        components: components
                        count: 4];
}

+ (NSColor *) ppManagedColor_sRGBColorWithWhite: (CGFloat) white
                                        alpha: (CGFloat) alpha
{
    return [self ppManagedColor_sRGBColorWithRed: white
                                        green: white
                                        blue: white
                                        alpha: alpha];
}

- (NSColor *) ppManagedColor_sRGBColor
{
    static NSColorSpace *colorSpace_sRGB = nil;
    NSColor *sRGBColor;

    if (!colorSpace_sRGB)
    {
        colorSpace_sRGB = [[NSColorSpace ppColorSpace_sRGB] retain];
    }

    sRGBColor = [self colorUsingColorSpace: colorSpace_sRGB];

    if (!sRGBColor)
        goto ERROR;

    return sRGBColor;

ERROR:
    return [[self retain] autorelease];
}

@end

@implementation NSBitmapImageRep (PPManagedColor)

- (NSColorSpace *) ppManagedColor_CalibratedColorSpace
{
    static bool needToCheckRespondsToColorSpaceSelector = YES, respondsToColorSpaceSelector = NO;
    NSColorSpace *colorSpace = nil;

    if (needToCheckRespondsToColorSpaceSelector)
    {
        // -[NSBitmapImageRep colorSpace] available on OS X 10.6 & later

        respondsToColorSpaceSelector =
            [NSBitmapImageRep instancesRespondToSelector: @selector(colorSpace)] ? YES : NO;

        needToCheckRespondsToColorSpaceSelector = NO;
    }

    if (respondsToColorSpaceSelector)
    {
        colorSpace = [self performSelector: @selector(colorSpace)];
    }
    else if ([self ppIsImageBitmap])
    {
        colorSpace = [NSColorSpace ppColorSpace_sRGB];
    }

    return colorSpace;
}

- (void) ppManagedColor_SetCalibratedColorSpace_sRGB
{
    static NSData *iccProfile_sRGB = nil;

    if (!iccProfile_sRGB)
    {
        iccProfile_sRGB = [[[NSColorSpace ppColorSpace_sRGB] ICCProfileData] retain];
    }

    if (iccProfile_sRGB)
    {
        [self setProperty: NSImageColorSyncProfileData withValue: iccProfile_sRGB];
    }
}

- (void) ppManagedColor_SetCalibratedColorSpaceFromBitmap: (NSBitmapImageRep *) bitmap
{
    NSData *iccProfile = [bitmap valueForProperty: NSImageColorSyncProfileData];

    if (iccProfile)
    {
        [self setProperty: NSImageColorSyncProfileData withValue: iccProfile];
    }
}

@end

#pragma mark Private methods

@implementation NSColorSpace (PPManagedColorPrivateMethods)

+ (NSColorSpace *) ppColorSpace_sRGB
{
    static NSColorSpace *sRGBColorSpace = nil;

    if (!sRGBColorSpace)
    {
        if ([[NSColorSpace class] respondsToSelector: @selector(sRGBColorSpace)])
        {
            sRGBColorSpace =
                [[[NSColorSpace class] performSelector: @selector(sRGBColorSpace)] retain];
        }

        if (!sRGBColorSpace)
        {
            NSData *iccProfile_sRGB = [NSData ppICCProfileFromSystemLibraryFolder_sRGB];

            if (iccProfile_sRGB)
            {
                sRGBColorSpace = [[NSColorSpace alloc] initWithICCProfileData: iccProfile_sRGB];
            }
        }

        if (!sRGBColorSpace)
        {
            // can't get sRGB - just use generic RGB

            sRGBColorSpace = [[NSColorSpace genericRGBColorSpace] retain];
        }
    }

    return sRGBColorSpace;
}

@end

@implementation NSData (PPManagedColorPrivateMethods)

#if defined(__APPLE__)

#   define kFilepath_ICCProfile_sRGB    @"/System/Library/ColorSync/Profiles/sRGB Profile.icc"

#else   // !defined(__APPLE__)

#   define kFilepath_ICCProfile_sRGB    nil

#endif  // !defined(__APPLE__)


+ (NSData *) ppICCProfileFromSystemLibraryFolder_sRGB
{
    NSData *iccProfile = nil;
    NSString *iccProfileFilepath = kFilepath_ICCProfile_sRGB;

    if (iccProfileFilepath)
    {
        iccProfile = [NSData dataWithContentsOfFile: iccProfileFilepath];
    }

    return iccProfile;
}

@end

#else   // !PP_DEPLOYMENT_TARGET_SUPPORTS_COLOR_MANAGEMENT

#pragma mark Managed Color Unsupported

@implementation NSColor (PPManagedColor)

+ (NSColor *) ppManagedColor_sRGBColorWithRed: (CGFloat) red
                                        green: (CGFloat) green
                                        blue: (CGFloat) blue
                                        alpha: (CGFloat) alpha
{
    return [self colorWithCalibratedRed: red green: green blue: blue alpha: alpha];
}

+ (NSColor *) ppManagedColor_sRGBColorWithWhite: (CGFloat) white
                                        alpha: (CGFloat) alpha
{
    return [self ppManagedColor_sRGBColorWithRed: white
                                        green: white
                                        blue: white
                                        alpha: alpha];
}

- (NSColor *) ppManagedColor_sRGBColor
{
    return [self colorUsingColorSpace: [NSColorSpace genericRGBColorSpace]];
}

@end

@implementation NSBitmapImageRep (PPManagedColor)

- (NSColorSpace *) ppManagedColor_CalibratedColorSpace
{
    return [NSColorSpace genericRGBColorSpace];
}

- (void) ppManagedColor_SetCalibratedColorSpace_sRGB
{
}

- (void) ppManagedColor_SetCalibratedColorSpaceFromBitmap: (NSBitmapImageRep *) bitmap
{
}

@end

#endif  // !PP_DEPLOYMENT_TARGET_SUPPORTS_COLOR_MANAGEMENT
